==> libeufin-0.12.0-sources.tar.gz: libeufin-0.12.0-sources/bank/src/main/kotlin/tech/libeufin/bank/api/WireGatewayApi.kt <==
/*
 * This file is part of LibEuFin.
 * Copyright (C) 2023-2024 Taler Systems S.A.

 * LibEuFin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation; either version 3, or
 * (at your option) any later version.

 * LibEuFin is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
 * Public License for more details.

 * You should have received a copy of the GNU Affero General Public
 * License along with LibEuFin; see the file COPYING.  If not, see
 * <http://www.gnu.org/licenses/>
 */

// This file contains the Taler Wire Gateway API handlers.

package tech.libeufin.bank.api

import io.ktor.http.*
import io.ktor.server.application.*
import io.ktor.server.request.*
import io.ktor.server.response.*
import io.ktor.server.routing.*
import io.ktor.util.pipeline.*
import tech.libeufin.bank.*
import tech.libeufin.bank.auth.auth
import tech.libeufin.bank.auth.authAdmin
import tech.libeufin.bank.auth.username
import tech.libeufin.bank.db.Database
import tech.libeufin.bank.db.ExchangeDAO
import tech.libeufin.bank.db.ExchangeDAO.AddIncomingResult
import tech.libeufin.bank.db.ExchangeDAO.TransferResult
import tech.libeufin.common.*
import java.time.Instant


fun Routing.wireGatewayApi(db: Database, ctx: BankConfig) {
    auth(db, TokenScope.readwrite) {
        get("/accounts/{USERNAME}/taler-wire-gateway/config") {
            call.respond(WireGatewayConfig(
                currency = ctx.regionalCurrency
            ))
        }
        post("/accounts/{USERNAME}/taler-wire-gateway/transfer") {
            val req = call.receive<TransferRequest>()
            ctx.checkRegionalCurrency(req.amount)
            val res = db.exchange.transfer(
                req = req,
                login = username,
                timestamp = Instant.now()
            )
            when (res) {
                is TransferResult.UnknownExchange -> throw unknownAccount(username)
                is TransferResult.NotAnExchange -> throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )
                is TransferResult.UnknownCreditor -> throw unknownCreditorAccount(req.credit_account.canonical)
                is TransferResult.BothPartyAreExchange -> throw conflict(
                    "Wire transfer attempted with credit and debit party being both exchange account",
                    TalerErrorCode.BANK_ACCOUNT_IS_EXCHANGE
                )
                is TransferResult.ReserveUidReuse -> throw conflict(
                    "request_uid used already",
                    TalerErrorCode.BANK_TRANSFER_REQUEST_UID_REUSED
                )
                is TransferResult.BalanceInsufficient -> throw conflict(
                    "Insufficient balance for exchange",
                    TalerErrorCode.BANK_UNALLOWED_DEBIT
                )
                is TransferResult.Success -> call.respond(
                    TransferResponse(
                        timestamp = res.timestamp,
                        row_id = res.id
                    )
                )
            }
        }
    }
    auth(db, TokenScope.readonly) {
        suspend fun <T> PipelineContext<Unit, ApplicationCall>.historyEndpoint(
            reduce: (List<T>, String) -> Any, 
            dbLambda: suspend ExchangeDAO.(HistoryParams, Long, BankPaytoCtx) -> List<T>
        ) {
            val params = HistoryParams.extract(context.request.queryParameters)
            val bankAccount = call.bankInfo(db, ctx.payto)
            
            if (!bankAccount.isTalerExchange)
                throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )

            val items = db.exchange.dbLambda(params, bankAccount.bankAccountId, ctx.payto)

            if (items.isEmpty()) {
                call.respond(HttpStatusCode.NoContent)
            } else {
                call.respond(reduce(items, bankAccount.payto))
            }
        }
        get("/accounts/{USERNAME}/taler-wire-gateway/history/incoming") {
            historyEndpoint(::IncomingHistory, ExchangeDAO::incomingHistory)
        }
        get("/accounts/{USERNAME}/taler-wire-gateway/history/outgoing") {
            historyEndpoint(::OutgoingHistory, ExchangeDAO::outgoingHistory)
        }
    }
    authAdmin(db, TokenScope.readwrite) {
        post("/accounts/{USERNAME}/taler-wire-gateway/admin/add-incoming") {
            val req = call.receive<AddIncomingRequest>()
            ctx.checkRegionalCurrency(req.amount)
            val timestamp = Instant.now()
            val res = db.exchange.addIncoming(
                req = req,
                login = username,
                timestamp = timestamp
            )
            when (res) {
                is AddIncomingResult.UnknownExchange -> throw unknownAccount(username)
                is AddIncomingResult.NotAnExchange -> throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )
                is AddIncomingResult.UnknownDebtor -> throw conflict(
                    "Debtor account ${req.debit_account} was not found",
                    TalerErrorCode.BANK_UNKNOWN_DEBTOR
                )
                is AddIncomingResult.BothPartyAreExchange -> throw conflict(
                    "Wire transfer attempted with credit and debit party being both exchange account",
                    TalerErrorCode.BANK_ACCOUNT_IS_EXCHANGE
                )
                is AddIncomingResult.ReservePubReuse -> throw conflict(
                    "reserve_pub used already",
                    TalerErrorCode.BANK_DUPLICATE_RESERVE_PUB_SUBJECT
                )
                is AddIncomingResult.BalanceInsufficient -> throw conflict(
                    "Insufficient balance for debitor",
                    TalerErrorCode.BANK_UNALLOWED_DEBIT
                )
                is AddIncomingResult.Success -> call.respond(
                    AddIncomingResponse(
                        timestamp = TalerProtocolTimestamp(timestamp),
                        row_id = res.id
                    )
                )
            }
        }
    }
}

==> libeufin-0.13.0-sources.tar.gz: libeufin-0.13.0-sources/bank/src/main/kotlin/tech/libeufin/bank/api/WireGatewayApi.kt <==
/*
 * This file is part of LibEuFin.
 * Copyright (C) 2023-2024 Taler Systems S.A.

 * LibEuFin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation; either version 3, or
 * (at your option) any later version.

 * LibEuFin is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General
 * Public License for more details.

 * You should have received a copy of the GNU Affero General Public
 * License along with LibEuFin; see the file COPYING.  If not, see
 * <http://www.gnu.org/licenses/>
 */

// This file contains the Taler Wire Gateway API handlers.

package tech.libeufin.bank.api

import io.ktor.http.*
import io.ktor.server.application.*
import io.ktor.server.request.*
import io.ktor.server.response.*
import io.ktor.server.routing.*
import io.ktor.util.pipeline.*
import tech.libeufin.bank.*
import tech.libeufin.bank.auth.auth
import tech.libeufin.bank.auth.authAdmin
import tech.libeufin.bank.auth.username
import tech.libeufin.bank.db.Database
import tech.libeufin.bank.db.ExchangeDAO
import tech.libeufin.bank.db.ExchangeDAO.AddIncomingResult
import tech.libeufin.bank.db.ExchangeDAO.TransferResult
import tech.libeufin.common.*
import java.time.Instant


fun Routing.wireGatewayApi(db: Database, ctx: BankConfig) {
    auth(db, ctx.pwCrypto, TokenScope.readwrite) {
        get("/accounts/{USERNAME}/taler-wire-gateway/config") {
            call.respond(WireGatewayConfig(
                currency = ctx.regionalCurrency
            ))
        }
        post("/accounts/{USERNAME}/taler-wire-gateway/transfer") {
            val req = call.receive<TransferRequest>()
            ctx.checkRegionalCurrency(req.amount)
            val res = db.exchange.transfer(
                req = req,
                login = username,
                timestamp = Instant.now()
            )
            when (res) {
                is TransferResult.UnknownExchange -> throw unknownAccount(username)
                is TransferResult.NotAnExchange -> throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )
                is TransferResult.UnknownCreditor -> throw unknownCreditorAccount(req.credit_account.canonical)
                is TransferResult.BothPartyAreExchange -> throw conflict(
                    "Wire transfer attempted with credit and debit party being both exchange account",
                    TalerErrorCode.BANK_ACCOUNT_IS_EXCHANGE
                )
                is TransferResult.ReserveUidReuse -> throw conflict(
                    "request_uid used already",
                    TalerErrorCode.BANK_TRANSFER_REQUEST_UID_REUSED
                )
                is TransferResult.BalanceInsufficient -> throw conflict(
                    "Insufficient balance for exchange",
                    TalerErrorCode.BANK_UNALLOWED_DEBIT
                )
                is TransferResult.Success -> call.respond(
                    TransferResponse(
                        timestamp = res.timestamp,
                        row_id = res.id
                    )
                )
            }
        }
    }
    auth(db, ctx.pwCrypto, TokenScope.readonly) {
        suspend fun <T> PipelineContext<Unit, ApplicationCall>.historyEndpoint(
            reduce: (List<T>, String) -> Any, 
            dbLambda: suspend ExchangeDAO.(HistoryParams, Long, BankPaytoCtx) -> List<T>
        ) {
            val params = HistoryParams.extract(context.request.queryParameters)
            val bankAccount = call.bankInfo(db, ctx.payto)
            
            if (!bankAccount.isTalerExchange)
                throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )

            val items = db.exchange.dbLambda(params, bankAccount.bankAccountId, ctx.payto)

            if (items.isEmpty()) {
                call.respond(HttpStatusCode.NoContent)
            } else {
                call.respond(reduce(items, bankAccount.payto))
            }
        }
        get("/accounts/{USERNAME}/taler-wire-gateway/history/incoming") {
            historyEndpoint(::IncomingHistory, ExchangeDAO::incomingHistory)
        }
        get("/accounts/{USERNAME}/taler-wire-gateway/history/outgoing") {
            historyEndpoint(::OutgoingHistory, ExchangeDAO::outgoingHistory)
        }
        get("/accounts/{USERNAME}/taler-wire-gateway/transfers") {
            val params = TransferParams.extract(context.request.queryParameters)
            val bankAccount = call.bankInfo(db, ctx.payto)
            if (!bankAccount.isTalerExchange)
                throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )
            
            if (params.status != null && params.status != TransferStatusState.success) {
                call.respond(HttpStatusCode.NoContent)
            } else {
                val items = db.exchange.pageTransfer(params.page, bankAccount.bankAccountId, ctx.payto)

                if (items.isEmpty()) {
                    call.respond(HttpStatusCode.NoContent)
                } else {
                    call.respond(TransferList(items, bankAccount.payto))
                }
            }
        }
        get("/accounts/{USERNAME}/taler-wire-gateway/transfers/{ROW_ID}") {
            val bankAccount = call.bankInfo(db, ctx.payto)
            if (!bankAccount.isTalerExchange)
                throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )
            
            val txId = call.longPath("ROW_ID")
            val transfer = db.exchange.getTransfer(bankAccount.bankAccountId, txId, ctx.payto) ?: throw notFound(
                "Transfer '$txId' not found",
                TalerErrorCode.BANK_TRANSACTION_NOT_FOUND
            )
            call.respond(transfer)
        }
    }
    authAdmin(db, ctx.pwCrypto, TokenScope.readwrite) {
        suspend fun ApplicationCall.addIncoming(
            amount: TalerAmount,
            debitAccount: Payto,
            subject: String,
            metadata: TalerIncomingMetadata
        ) {
            ctx.checkRegionalCurrency(amount)
            val timestamp = Instant.now()
            val res = db.exchange.addIncoming(
                amount = amount,
                debitAccount = debitAccount,
                subject = subject,
                login = username,
                timestamp = timestamp,
                metadata = metadata
            )
            when (res) {
                is AddIncomingResult.UnknownExchange -> throw unknownAccount(username)
                is AddIncomingResult.NotAnExchange -> throw conflict(
                    "$username is not an exchange account.",
                    TalerErrorCode.BANK_ACCOUNT_IS_NOT_EXCHANGE
                )
                is AddIncomingResult.UnknownDebtor -> throw conflict(
                    "Debtor account $debitAccount was not found",
                    TalerErrorCode.BANK_UNKNOWN_DEBTOR
                )
                is AddIncomingResult.BothPartyAreExchange -> throw conflict(
                    "Wire transfer attempted with credit and debit party being both exchange account",
                    TalerErrorCode.BANK_ACCOUNT_IS_EXCHANGE
                )
                is AddIncomingResult.ReservePubReuse -> throw conflict(
                    "reserve_pub used already",
                    TalerErrorCode.BANK_DUPLICATE_RESERVE_PUB_SUBJECT
                )
                is AddIncomingResult.BalanceInsufficient -> throw conflict(
                    "Insufficient balance for debitor",
                    TalerErrorCode.BANK_UNALLOWED_DEBIT
                )
                is AddIncomingResult.Success -> this.respond(
                    AddIncomingResponse(
                        timestamp = TalerProtocolTimestamp(timestamp),
                        row_id = res.id
                    )
                )
            }
        }
        post("/accounts/{USERNAME}/taler-wire-gateway/admin/add-incoming") {
            val req = call.receive<AddIncomingRequest>()
            call.addIncoming(
                amount = req.amount,
                debitAccount = req.debit_account,
                subject = "Manual incoming ${req.reserve_pub}",
                metadata = TalerIncomingMetadata(TalerIncomingType.reserve, req.reserve_pub)
            )
        }
        post("/accounts/{USERNAME}/taler-wire-gateway/admin/add-kycauth") {
            val req = call.receive<AddKycauthRequest>()
            call.addIncoming(
                amount = req.amount,
                debitAccount = req.debit_account,
                subject = "Manual incoming KYC:${req.account_pub}",
                metadata = TalerIncomingMetadata(TalerIncomingType.kyc, req.account_pub)
            )
        }
    }
}
